
import torch
import torch.nn as nn

class UNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv2d = nn.Conv2d(1, 64, kernel_size=3, padding='same', padding_mode='zeros')
        self.conv2d_1 = nn.Conv2d(64, 64, kernel_size=3, padding='same')
        self.conv2d_2 = nn.Conv2d(64, 128, kernel_size=3, padding='same')
        self.conv2d_3 = nn.Conv2d(128, 128, kernel_size=3, padding='same')
        self.conv2d_4 = nn.Conv2d(128, 256, kernel_size=3, padding='same')
        self.conv2d_5 = nn.Conv2d(256, 256, kernel_size=3, padding='same')
        self.conv2d_6 = nn.Conv2d(256, 512, kernel_size=3, padding='same')
        self.conv2d_7 = nn.Conv2d(512, 512, kernel_size=3, padding='same')
        self.conv2d_8 = nn.Conv2d(512, 1024, kernel_size=3, padding='same')
        self.conv2d_9 = nn.Conv2d(1024, 1024, kernel_size=3, padding='same')
        self.conv2d_10 = nn.Conv2d(1024, 512, kernel_size=2, padding='same')
        self.conv2d_11 = nn.Conv2d(1024, 512, kernel_size=3, padding='same')
        self.conv2d_12 = nn.Conv2d(512, 512, kernel_size=3, padding='same')
        self.conv2d_13 = nn.Conv2d(512, 256, kernel_size=2, padding='same')
        self.conv2d_14 = nn.Conv2d(512, 256, kernel_size=3, padding='same')
        self.conv2d_15 = nn.Conv2d(256, 256, kernel_size=3, padding='same')
        self.conv2d_16 = nn.Conv2d(256, 128, kernel_size=2, padding='same')
        self.conv2d_17 = nn.Conv2d(256, 128, kernel_size=3, padding='same')
        self.conv2d_18 = nn.Conv2d(128, 128, kernel_size=3, padding='same')
        self.conv2d_19 = nn.Conv2d(128, 64, kernel_size=2, padding='same')
        self.conv2d_20 = nn.Conv2d(128, 64, kernel_size=3, padding='same')
        self.conv2d_21 = nn.Conv2d(64, 64, kernel_size=3, padding='same')
        self.conv2d_22 = nn.Conv2d(64, 2, kernel_size=3, padding='same')
        self.conv2d_23 = nn.Conv2d(2, 1, kernel_size=1, padding='same')

        # Other stuff
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d((2,2), stride=(2,2))
        self.dropout = nn.Dropout(0.5)
        self.upsample =  nn.Upsample(scale_factor=2, mode='nearest')
        self.sigmoid = nn.Sigmoid()

    def forward(self, inputs):
        """ Encoder """
        conv1 = self.conv2d(inputs)
        conv1 = self.relu(conv1)
        conv1 = self.conv2d_1(conv1)
        conv1 = self.relu(conv1)
        pool1 = self.maxpool(conv1)

        conv2 = self.conv2d_2(pool1)
        conv2 = self.relu(conv2)
        conv2 = self.conv2d_3(conv2)
        conv2= self.relu(conv2)
        pool2 = self.maxpool(conv2)

        conv3 = self.conv2d_4(pool2)
        conv3 = self.relu(conv3)
        conv3 = self.conv2d_5(conv3)
        conv3 = self.relu(conv3)
        pool3 = self.maxpool(conv3)

        conv4 = self.conv2d_6(pool3)
        conv4 = self.relu(conv4)
        conv4 = self.conv2d_7(conv4)
        conv4 = self.relu(conv4)
        drop4 = self.dropout(conv4)
        pool4 = self.maxpool(drop4)

        
        """ Bottleneck """
        conv5 = self.conv2d_8(pool4)
        conv5 = self.relu(conv5)
        conv5 = self.conv2d_9(conv5)
        conv5 = self.relu(conv5)
        drop5 = self.dropout(conv5)

        """ Decoder """
        up6 = self.upsample(drop5)
        up6 = self.conv2d_10(up6)
        up6 = self.relu(up6)
        merg6 = torch.cat((drop4,up6), dim=1)
        conv6 = self.conv2d_11(merg6)
        conv6 = self.relu(conv6)
        conv6 = self.conv2d_12(conv6)
        conv6 = self.relu(conv6)

        up7 = self.upsample(conv6)
        up7 = self.conv2d_13(up7)
        up7 = self.relu(up7)
        merg7 = torch.cat((conv3,up7), dim=1)
        conv7 = self.conv2d_14(merg7)
        conv7 = self.relu(conv7)
        conv7 = self.conv2d_15(conv7)
        conv7 = self.relu(conv7)

        up8 = self.upsample(conv7)
        up8 = self.conv2d_16(up8)
        up8 = self.relu(up8)
        merg8 = torch.cat((conv2,up8), dim=1)
        conv8 = self.conv2d_17(merg8)
        conv8 = self.relu(conv8)
        conv8 = self.conv2d_18(conv8)
        conv8 = self.relu(conv8)

        up9 = self.upsample(conv8)
        up9 = self.conv2d_19(up9)
        up9 = self.relu(up9)
        merg9 = torch.cat((conv1,up9), dim=1)
        conv9 = self.conv2d_20(merg9)
        conv9 = self.relu(conv9)
        conv9 = self.conv2d_21(conv9)
        conv9 = self.relu(conv9)

        """ Classifier """
        conv9 = self.conv2d_22(conv9)
        conv9 = self.relu(conv9)
        conv10 = self.conv2d_23(conv9)
        conv10 = self.sigmoid(conv10)
        
        
        return conv10





# -*- coding: utf-8 -*-
"""
Created on Sat Dec 21 18:54:10 2019

"""
import os
import sys
from .model_pytorch import UNet
# from model_tensorflow import unet
import numpy as np
import skimage
from skimage import io
import skimage.transform as trans
import torch

import importlib.resources as resources
# Access weight file
path_weights = resources.files('yeaz.unet.weights')

def create_directory_if_not_exists(path):
    """
    Create in the file system a new directory if it doesn't exist yet.
    Param:
        path: the path of the new directory
    """
    if not os.path.exists(path):
        os.makedirs(path)


def threshold(im,th = None):
    """
    Binarize an image with a threshold given by the user, or if the threshold is None, calculate the better threshold with isodata
    Param:
        im: a numpy array image (numpy array)
        th: the value of the threshold (feature to select threshold was asked by the lab)
    Return:
        bi: threshold given by the user (numpy array)
    """
    im2 = im.copy()
    if th == None:
        th = skimage.filters.threshold_isodata(im2)
    bi = im2
    bi[bi > th] = 255
    bi[bi <= th] = 0
    return bi

import gc
def report_gpu():
    gc.collect()
    torch.cuda.empty_cache()
    print('Memory allocated: ', torch.cuda.memory_allocated())

def prediction(im, mic_type, pretrained_weights=None, model_type='pytorch', device=None):
    """
    Calculate the prediction of the label corresponding to image im
    Param:
        im: a numpy array image (numpy array), with max size 2048x2048
    Return:
        res: the predicted distribution of probability of the labels (numpy array)
    """        
    # pad with zeros such that is divisible by 16
    (nrow, ncol) = im.shape
    row_add = 16-nrow%16
    col_add = 16-ncol%16
    padded = np.pad(im, ((0, row_add), (0, col_add)))
    
    if pretrained_weights is None:
        if mic_type == 'pc':
            pretrained_weights = str(path_weights) + '/' + 'weights_budding_PhC_multilab_0_1'
        elif mic_type == 'bf':
            pretrained_weights = str(path_weights) + '/' + 'weights_budding_BF_multilab_0_1'
        elif mic_type == 'fission':
            pretrained_weights = str(path_weights) + '/' + 'weights_fission_multilab_0_2'
        if model_type == 'tensorflow':
            pretrained_weights = pretrained_weights + '.hdf5'
    
    if not os.path.exists(pretrained_weights):
        raise ValueError('Path does not exist')

    # WHOLE CELL PREDICTION

    if model_type == 'tensorflow':
        tf_model = unet(pretrained_weights = pretrained_weights,
                    input_size = (None,None,1))
        input = padded[np.newaxis,:,:,np.newaxis]
        tf_results = tf_model.predict(input, batch_size=1)
        tf_res = tf_results[0,:,:,0]
        
        return tf_res

    elif model_type == 'pytorch':
        # Load saved weights in pytorch model and run the pytorch model
        model = UNet()
        model.load_state_dict(torch.load(pretrained_weights))
        
        if torch.cuda.is_available() and device == 'cuda':
            device = torch.device('cuda')
            model = model.to(device)
            padded = torch.from_numpy(padded).to(device)
            print('device: ', device)
        else:
            device = torch.device('cpu')
            padded = torch.from_numpy(padded)
            print('device: ', device)
        model.eval()
        with torch.no_grad():
            # Convert input tensor to PyTorch tensor
            input_tensor = padded.unsqueeze(0).unsqueeze(0).float()
            # Pass input through the model
            output_tensor = model.forward(input_tensor)
            # Convert output tensor to NumPy array
            output_array = output_tensor.cpu().detach().numpy()
        pt_res = output_array[0, 0, :, :]
        
        # empty cache
        if (device == 'cuda') and (torch.cuda.is_available()):
            report_gpu()
        
        return pt_res[:nrow, :ncol]
    
    
    else:
        raise ValueError('model_type is not valid. should be either "pytorch" or "tensorflow".')





def _setup_gui():
    try:
        from PyQt6.QtCore import Qt
    except Exception as e:
        txt = (
            'To run the GUI YeaZ needs to install a package called `PyQt6`.\n\n'
            'You can let YeaZ install it now, or you can abort (press "n")\n'
            'and install it manually with the following command:\n\n'
            'pip install PyQt6\n'
        )
        print('-'*60)
        print(txt)
        while True:
            answer = input('Do you want to install PyQt6 now ([y]/n)? ')
            if answer.lower() == 'y' or not answer:
                import subprocess
                import sys
                subprocess.check_call(
                    [sys.executable, '-m', 'pip', 'install', '-U', 'PyQt6']
                )
                warn_restart = True
                break
            elif answer.lower() == 'n':
                raise e
            else:
                print(
                    f'"{answer}" is not a valid answer. '
                    'Type "y" for "yes", or "n" for "no".'
                )

def run():
    _setup_gui()
    
    print('Initializing application...')
    
    import sys
    from PyQt6.QtWidgets import QApplication
    
    from .GUI_main import App
    #this file contains a dialog window which is opened before the main program
    #and allows to load the nd2 and hdf files by browsing through the computer.
    from .disk import DialogFileBrowser as dfb
    
    app = QApplication(sys.argv)
    
    # If two arguments are given, make them nd2name and hdfname
    if len(sys.argv)==3:
        nd2name1 = sys.argv[1]
        hdfname1 = sys.argv[2]
        ex = App(nd2name1, hdfname1, '')
        print('------------------------------------------------ Welcome to YeaZ-GUI -------------------------------------------------------')

        sys.exit(app.exec())
    
    # Launch file browser otherwise
    else:
        wind = dfb.FileBrowser()
        if wind.exec():
            nd2name1 = wind.nd2name 
            hdfname1 = wind.hdfname
            hdfnewname = wind.newhdfname

            ex = App(str(nd2name1), str(hdfname1), str(hdfnewname))
            print('------------------------------------------------ Welcome to YeaZ-GUI -------------------------------------------------------')
            sys.exit(app.exec())
        else:
            app.exit()


# -*- coding: utf-8 -*-
"""
Initializing the layout of the main window. It places the buttons and the 
pictures at the desired positions.
"""
from matplotlib.backends.qt_compat import QtWidgets


def Init(parent):

    
    # LAYOUT OF THE MAIN WINDOW
    layout = QtWidgets.QVBoxLayout(parent._main)
    
    # LAYOUT FOR THE TOOLBAR BUTTONS
    hbox = QtWidgets.QHBoxLayout()
    hbox.addWidget(parent.button_home) 
    hbox.addWidget(parent.button_back)
    hbox.addWidget(parent.button_forward)
    hbox.addWidget(parent.button_pan)
    hbox.addWidget(parent.button_zoom)
    hbox.addStretch(1)
    layout.addLayout(hbox)
    
    # TIME NAVIGATION
    hboxtimeframes = QtWidgets.QHBoxLayout()
    hboxtimeframes.addWidget(parent.button_fov)
    hboxtimeframes.addWidget(parent.button_channel)
    hboxtimeframes.addStretch()
    hboxtimeframes.addWidget(parent.button_previousframe)
    hboxtimeframes.addWidget(parent.button_timeindex)
    hboxtimeframes.addWidget(parent.button_nextframe)
    hboxtimeframes.addStretch()
    
    # IMAGE DISPLAY
    hlayout = QtWidgets.QHBoxLayout()
    hlayout.addWidget(parent.m)
    hlayout.setContentsMargins(0, 0, 0, 0)
    layout.addLayout(hlayout)        
    layout.addLayout(hboxtimeframes)
    
    # CORRECTIONS/ IMAGE EDITS       
    hboxcorrectionsbuttons = QtWidgets.QHBoxLayout()
    hboxcorrectionsbuttons.addWidget(parent.button_add_region)
    hboxcorrectionsbuttons.addWidget(parent.button_newcell)
    hboxcorrectionsbuttons.addWidget(parent.button_split)
    hboxcorrectionsbuttons.addWidget(parent.button_mergewithneighbors)
    hboxcorrectionsbuttons.addWidget(parent.button_drawmouse)
    hboxcorrectionsbuttons.addWidget(parent.button_eraser)
    hboxcorrectionsbuttons.addWidget(parent.label_brushsize)
    hboxcorrectionsbuttons.addWidget(parent.spinbox_brushsize)
    hboxcorrectionsbuttons.addStretch(2)
    hboxcorrectionsbuttons.addWidget(parent.button_showval)
    hboxcorrectionsbuttons.addWidget(parent.button_hidemask)
    layout.addLayout(hboxcorrectionsbuttons)
    
    # LAYOUT FOR EDITING CELL VALUES, SHOW CELL VALUES AND HIDE MASK, CNN BUTTONS
    hboxcellval = QtWidgets.QHBoxLayout()
    hboxcellval.addWidget(parent.button_exval)
    hboxcellval.addWidget(parent.button_changecellvalue)
    hboxcellval.addWidget(parent.button_cellcorrespondence)
    hboxcellval.addStretch(1)
    hboxcellval.addWidget(parent.button_cnn)
    hboxcellval.addWidget(parent.button_extractfluorescence)
    layout.addLayout(hboxcellval)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 15 17:05:00 2020

@mattminder
"""

import os
import re
from skimage import io
import numpy as np


def load_image(path, ix=None):
    """Loads Image at specified path. Path can be to single image as supported
    by skimage.io, or to folder containing images supported by skimage.io.
    Ix specifies which image should be loaded, if None it loads all images."""
    
    _, ext = os.path.splitext(path)
    
    # Folder
    if ext=='':
        filelist = sorted(os.listdir(path)) 
        filelist = [f for f in filelist if 
                    re.search(r".png|.tif|.jpg|.bmp|.jpeg|.pbm|.pgm|.ppm|.pxm|.pnm|.jp2|.PNG|.TIF|.JPG|.BMP|.JPEG|.PBM|.PGM|.PPM|.PXM|.PNM|.JP2", f)]
        filelist = [os.path.join(path, f) for f in filelist]
        
        if len(filelist)==0:
            raise ValueError('Folder does not contain images')
        
        if ix is None:
            ims = [io.imread(f) for f in filelist]
            ims = np.array(ims)
            return ims
        else:
            im = io.imread(filelist[ix])
            return im
        
    # File
    else:
        try:
            im = io.imread(path)
        except ValueError:
            raise ValueError('Not an image file')
        
        # Single image
        if im.ndim == 2:
            if ix is not None:
                return im 
            else:
                return im[None,:,:]
        
        # Multistack image
        if im.ndim == 3:
            if ix is None:
                return im 
            else: 
                return im[ix,:,:]
        




from PyQt6.QtWidgets import QApplication, QMainWindow, QMenu, QVBoxLayout, QSizePolicy, QMessageBox, QWidget, QPushButton, QComboBox, QDialog, QDialogButtonBox, QInputDialog, QLineEdit, QFormLayout, QLabel
from PyQt6 import QtGui
from PyQt6.QtGui import QShortcut
from PyQt6.QtCore import pyqtSignal, QObject, Qt
#import PyQt package, allows for GUI interactions

class CustomDialog(QDialog):

    def __init__(self, *args, **kwargs):
        super(CustomDialog, self).__init__(*args, **kwargs)
        app, = args
        self.setWindowTitle("Retrack")
        self.setGeometry(100,100, 500,200)
        
        self.entry1 = QLineEdit()
        self.entry1.setValidator(QtGui.QIntValidator())
        self.entry1.setMaxLength(4)
        self.entry1.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.labeltime = QLabel("Enter a frame number between {} to {}".format(app.Tindex+1, app.reader.sizet-1))
        flo = QFormLayout()
        flo.addWidget(self.labeltime)
        flo.addRow('retracking frames from {} (next frame) to '.format(app.Tindex+1), self.entry1)       
        
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        flo.addWidget(self.buttonBox)
        self.setLayout(flo)
        
        # Set focus to entry1
        self.entry1.setFocus()
        



from model_pytorch import UNet
from model_tensorflow import unet
import torch
import tensorflow as tf

def weight_loading(pretrained_weights):
    # Load the weights
    tf_model = tf.keras.models.load_model(pretrained_weights)
    tf_weights = tf_model.get_weights()

    # Load the PyTorch model
    pt_model = UNet() #implemented based on the previous model (by myself)
    initial_state_dict = pt_model.state_dict()
    new_state_dict = {}
    with torch.no_grad():
        x = 0
        for i, layer in enumerate(pt_model.modules()):
            if isinstance(layer, torch.nn.Conv2d) or isinstance(layer, torch.nn.Linear):
                # extract the weights and biases from the TensorFlow weights
                weight_tf = tf_weights[x*2]
                bias_tf = tf_weights[x*2+1]

                # convert the weights and biases to PyTorch format
                layer.weight.data = torch.tensor(weight_tf.transpose(3, 2, 0, 1), dtype=torch.float)
                layer.bias.data = torch.tensor(bias_tf, dtype=torch.float)

                x = x + 1

            if isinstance(layer, torch.nn.ConvTranspose2d):
                weight_tf = tf_weights[x*2]
                bias_tf = tf_weights[x*2+1]
                
                # convert the weights and biases to PyTorch format
                layer.weight.data = torch.tensor(weight_tf.transpose(2, 3, 0, 1), dtype=torch.float)
                layer.bias.data = torch.tensor(bias_tf, dtype=torch.float)
                
                x = x + 1
    return pt_model
